﻿Option Strict On
Option Infer On

Imports SlimDX
Imports SlimDX.Direct2D
Imports SlimDX.DirectWrite
Imports D2DControlset
Imports Helpers.Mathematics
Imports Helpers.Direct2D
Imports System.Threading

Friend MustInherit Class GameControl
    Inherits D2DControl

    Private borderFactor As Single

    Protected ReadOnly Property outRect As RectangleF
        Get
            Dim bWidth As Single = borderWidth
            Dim bWidthH As Single = bWidth / 2.0F
            Dim aLoc As Point = AbsoluteLocation
            Dim rect As New RectangleF(aLoc.X + bWidthH, aLoc.Y + bWidthH, Width - bWidth, Height - bWidth)
            Return rect
        End Get
    End Property

    Protected ReadOnly Property borderWidth As Single
        Get
            Return borderFactor * Game.getSizeFactor
        End Get
    End Property

    Protected ReadOnly Property cornerSize As Single
        Get
            Return 10.0F * Game.getSizeFactor
        End Get
    End Property

    Protected Sub setBorderWidth(val As Single)
        borderFactor = val
    End Sub

    Private Function getGeometry(start As PointF, points() As PointF, filled As Boolean, e As D2DDrawEventArgs) As PathGeometry
        Dim path As New PathGeometry(e.Factory)
        Dim sink As GeometrySink = path.Open
        sink.BeginFigure(start, If(filled, FigureBegin.Filled, FigureBegin.Hollow))
        sink.AddLines(points)
        sink.EndFigure(FigureEnd.Closed)
        sink.Close()
        sink.Dispose()
        Return path
    End Function

    Protected Overrides Sub OnDraw(e As D2DDrawEventArgs)
        Dim bWidth As Single = borderWidth
        Dim bWidthH As Single = bWidth / 2.0F
        Dim outR As RectangleF = outRect
        Dim left As Single = outR.X
        Dim right As Single = outR.X + outR.Width
        Dim top As Single = outR.Y
        Dim bottom As Single = outR.Y + outR.Height
        Dim cSize As Single = cornerSize
        Dim startPoint As New PointF(left, top + cornerSize)
        Dim points(6) As PointF
        points(0) = New PointF(left + cornerSize, top)
        points(1) = New PointF(right - cornerSize, top)
        points(2) = New PointF(right, top + cornerSize)
        points(3) = New PointF(right, bottom - cornerSize)
        points(4) = New PointF(right - cornerSize, bottom)
        points(5) = New PointF(left + cornerSize, bottom)
        points(6) = New PointF(left, bottom - cornerSize)
        Dim white As New SolidColorBrush(e.Target, CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush).Color)
        Dim cWidthH = cornerSize / 2.0F
        Dim inRect As New RectangleF(outR.X + cWidthH, outR.Y + cWidthH, outR.Width - 2 * cWidthH, outR.Height - 2 * cWidthH)
        drawContent(e, inRect)
        Dim border As PathGeometry = getGeometry(startPoint, points, False, e)
        e.Target.DrawGeometry(border, white, bWidth)
        border.Dispose()
        white.Dispose()
    End Sub

    Protected MustOverride Sub drawContent(e As D2DDrawEventArgs, inRect As RectangleF)

    Sub New()
        MyBase.New()
        borderFactor = 4.0F
    End Sub

    Sub New(parent As IDirect2DControl)
        MyBase.New(parent)
        borderFactor = 4.0F
    End Sub
End Class

Friend Class Minimap
    Inherits GameControl

    Private world As World
    Private viewRect As Rectangle

    Friend Sub setView(rect As Rectangle)
        viewRect = rect
    End Sub

    Protected Overrides Sub drawContent(e As D2DDrawEventArgs, inRect As RectangleF)
        drawMap(e, inRect)
    End Sub

    Private Sub drawMap(e As D2DDrawEventArgs, inRect As RectangleF)
        Dim width As Single = 0.0F
        Dim height As Single = 0.0F
        Dim f As Single = 0.0F
        If world.size.Width >= world.size.Height Then
            f = inRect.Width / world.size.Width
        Else
            f = inRect.Height / world.size.Height
        End If
        width = world.size.Width * f
        height = world.size.Height * f
        Dim x As Single = (inRect.Width - width) / 2.0F + inRect.X
        Dim y As Single = (inRect.Height - height) / 2.0F + inRect.Y
        Dim poly(world.way.segments.Count - 1) As PointF
        Dim transformPoint As Func(Of PointF, PointF) = Function(p)
                                                            Dim pX As Single = p.X / world.size.Width * width
                                                            Dim pY As Single = p.Y / world.size.Height * height
                                                            Return New PointF(pX + x, pY + y)
                                                        End Function
        Dim path As New PathGeometry(e.Factory)
        Dim sink As GeometrySink = path.Open
        sink.BeginFigure(transformPoint(world.way.segments(0).p1), FigureBegin.Hollow)
        For i = 0 To world.way.segments.Length - 1
            sink.AddLine(transformPoint(world.way.segments(i).p2))
        Next
        sink.EndFigure(FigureEnd.Open)
        sink.Close()
        Dim b As New SolidColorBrush(e.Target, New Color4(Color.DimGray))
        Dim sProps As New StrokeStyleProperties
        sProps.LineJoin = LineJoin.Round
        sProps.StartCap = CapStyle.Round
        sProps.EndCap = CapStyle.Round
        Dim style As New StrokeStyle(e.Factory, sProps)
        e.Target.DrawGeometry(path, b, 24.0F * f, style)
        style.Dispose()
        b.Dispose()
        sink.Dispose()
        path.Dispose()
        Dim h As New Direct2DHelper(e.Factory, e.Target, e.TextFactory)
        Dim red As New SolidColorBrush(e.Target, New Color4(Color.Red))
        Dim green As New SolidColorBrush(e.Target, New Color4(Color.Lime))
        Dim ents() As Entity = world.getEntities
        For Each ent As Entity In ents
            Do While ent.InUse
                Thread.Sleep(1)
            Loop
            ent.InUse = True
            If TypeOf ent Is Creep Then
                Dim c As Creep = DirectCast(ent, Creep)
                Dim p As PointF = transformPoint(c.getLocation)
                h.FillCircle(New Helpers.Mathematics.Vector2(p), 12.0F * f, red)
            End If
            If TypeOf ent Is Tower Then
                Dim t As Tower = DirectCast(ent, Tower)
                Dim p As PointF = transformPoint(t.getLocation)
                h.FillCircle(New Helpers.Mathematics.Vector2(p), 14.0F * f, green)
            End If
            ent.InUse = False
        Next
        green.Dispose()
        red.Dispose()
        If viewRect.Width < world.size.Width OrElse viewRect.Height < world.size.Height Then e.Target.DrawRectangle(ColorManager.GetBrush("defaultWhite"), New RectangleF(transformPoint(viewRect.Location), New SizeF(viewRect.Width * f, viewRect.Height * f)))
    End Sub

    Sub New(ByVal w As World)
        world = w
    End Sub
End Class
