﻿Option Strict On
Option Infer On

Imports System.IO
Imports DirectShow
Imports System.Threading

Friend Class MusicManager
    Private Shared musics As List(Of Music)
    Private Shared env As MusicEnvironment
    Private Shared player As MediaPlayer
    Private Shared vol As Integer
    Private Shared t As Thread
    Private Shared exitThread As Boolean
    Private Shared envChanged As Boolean
    Private Const fadeTime = 500

    ''' <summary>
    ''' Die Lautstärke der Musik.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Friend Shared Property volume As Integer
        Get
            Return vol
        End Get
        Set(value As Integer)
            vol = value
        End Set
    End Property

    ''' <summary>
    ''' Die Spielumgebung, der die Musik angepasst wird.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Friend Shared Property environment As MusicEnvironment
        Get
            Return env
        End Get
        Set(value As MusicEnvironment)

            'bei gleicher Umgebung nichts tun
            If value = env Then Exit Property

            'neue Umgebung festlegen
            env = value
            envChanged = True

        End Set
    End Property

    ''' <summary>
    ''' Initialisiert das Musikmanagement.
    ''' </summary>
    ''' <remarks>Diese Methode muss aufgerufen werden, bevor Musik abgespielt werden kann.</remarks>
    Friend Shared Sub init()

        MusicEnvironment.init() 'Umgebungen initialisieren
        loadMusic() 'Musik laden

    End Sub

    ''' <summary>
    ''' Läd die verfügbare Musik und die passende Umgebung in das Programm.
    ''' </summary>
    ''' <remarks></remarks>
    Private Shared Sub loadMusic()

        'Einlesen
        Dim f As New FileInfo(Path.Combine(Game.getEnvironmentPath("musicconf"), "files.ini")) 'Dateipfad erstellen
        If Not f.Exists Then Throw New FileNotFoundException("Die Datei " & f.Name & " konnte nicht gefunden werden.", f.FullName) 'Datei nicht vorhanden
        musics = New List(Of Music) 'Liste initialisieren
        Dim s = String.Empty

        Using sr = File.OpenText(f.FullName) 'Datei öffnen

            'Zeilenweise einlesen
            Do Until sr.EndOfStream
                s &= sr.ReadLine
            Loop

        End Using


        'Auswerten
        Dim tokens() = s.Split({";"c}, StringSplitOptions.RemoveEmptyEntries) 'Inhalt in Tokens teilen

        For i = 0 To tokens.Length - 1 'Tokens iterieren

            Dim parts() = tokens(i).Split({","c}, 2) 'Token in Teile spalten
            Dim mF As New FileInfo(parts(0)) 'Dateipfad extrahieren
            Dim num As Integer
            Dim isNum = Integer.TryParse(parts(1), num) 'Umgebung extrahieren

            'zur Liste hinzufügen
            If mF.Exists AndAlso isNum AndAlso num < MusicEnvironment.environmentCount AndAlso num >= 0 Then 'auf Korrektheit prüfen

                'Musik-Struktur erstellen
                Dim m As New Music
                m.file = mF.FullName
                m.environment = MusicEnvironment.getEnvironment(num)
                musics.Add(m) 'hinzufügen

            End If

        Next

    End Sub

    ''' <summary>
    ''' Beginnt die Verwaltung der Musik.
    ''' </summary>
    ''' <remarks></remarks>
    Friend Shared Sub start()

        'Hintergrundthread erstellen
        exitThread = False
        t = New Thread(New ThreadStart(AddressOf threadLoop))
        t.Name = "music thread"
        t.IsBackground = True
        t.Start()

    End Sub

    ''' <summary>
    ''' Beendet die Verwaltung der Musik.
    ''' </summary>
    ''' <remarks></remarks>
    Friend Shared Sub [stop]()
        exitThread = True 'Hintergrundthread beenden
    End Sub

    ''' <summary>
    ''' Führt den Messageloop für die Musik aus.
    ''' </summary>
    ''' <remarks></remarks>
    Private Shared Sub threadLoop()

        player = New MediaPlayer 'Mediaplayer Initialisieren
        Dim stopping = False

        'Ereignisbehandlung hinzufügen
        AddHandler player.stopped, Sub()
                                       If player.state = MediaPlayerState.Stopped AndAlso env IsNot Nothing AndAlso Not stopping Then
                                           player.mediaFile = getNextMusicFile()
                                           player.play()
                                       End If
                                   End Sub

        'Messageloop
        Do Until exitThread 'bis Beendigung ausführen
            Thread.Sleep(10) 'Prozessorzeit sparen -> Latenz von 10ms

            If vol <> player.Volume Then player.Volume = vol 'Lautstärke setzen

            'wenn die Umgebung wechselt
            If envChanged Then
                envChanged = False

                'Fading vorbereiten
                Dim volumeStep = (vol - 1) / fadeTime
                Dim fade = True

                If player.state = MediaPlayerState.Playing Then

                    'Fade-Out
                    For i = 1 To fadeTime
                        Thread.Sleep(1)
                        player.Volume = vol - Convert.ToInt32(i * volumeStep)
                    Next

                    stopping = True
                    player.stop() 'Wiedergabe stoppen
                    stopping = False

                Else
                    fade = False 'kein Fade-In
                End If

                If env IsNot Nothing Then 'nur Musik abspielen, wenn neue Umgebung festgelegt

                    'Datei laden
                    player.mediaFile = getNextMusicFile()
                    player.play()

                    If fade Then 'Fade-In

                        For i = fadeTime - 1 To 0 Step -1 'rückwärts, weil von leise nach laut
                            Thread.Sleep(1)
                            player.Volume = vol - Convert.ToInt32(i * volumeStep)
                        Next

                    Else
                        player.Volume = vol 'sofort alte Lautstärke
                    End If

                End If

            End If

        Loop

        'vor Ende Mediaplayer stoppen
        If player.state <> MediaPlayerState.Stopped Then
            stopping = True
            player.stop()
            stopping = False
        End If

    End Sub

    ''' <summary>
    ''' gibt zufällig eine neu Mediendatei passen zur aktuellen Umgebung zurück.
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Shared Function getNextMusicFile() As String

        If env Is Nothing Then Return "" 'keine Umgebung gewählt

        Dim aM() = getMusicFiles() 'verfügbare Musik auflisten
        If aM.Length <= 0 Then Return "" 'keine Musik verfügbar

        'zufällige Musik auswählen
        Dim index = Game.rand.Next(aM.Length)
        Return aM(index)

    End Function

    ''' <summary>
    ''' Listet alle für die aktuelle Umgebung verfügbaren Mediendateien auf.
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Private Shared Function getMusicFiles() As String()

        Dim iEn As IEnumerable(Of String) = (From m In musics Where m.environment = env Select m.file) 'Linq für einfaches Suchen
        Return iEn.ToArray 'als Array zurückgeben

    End Function

    ''' <summary>
    ''' Musik-Struktur für einfachere Handhabung mit Umgebungen.
    ''' </summary>
    ''' <remarks></remarks>
    Private Structure Music
        Friend environment As MusicEnvironment 'Umgebung der Musik
        Friend file As String 'Mediendatei
    End Structure
End Class

Friend Class MusicEnvironment
    Private Shared environments As List(Of MusicEnvironment)

    Private m_Name As String
    Private m_ID As Integer

    Friend ReadOnly Property name As String
        Get
            Return m_Name
        End Get
    End Property

    Friend ReadOnly Property ID As Integer
        Get
            Return m_ID
        End Get
    End Property

    Friend Shared ReadOnly Property environmentCount As Integer
        Get
            Return environments.Count
        End Get
    End Property

    ''' <summary>
    ''' Initialisiert alle Musikumgebungen.
    ''' </summary>
    ''' <remarks></remarks>
    Friend Shared Sub init()

        'Einlesen
        Dim f = Path.Combine(Game.getEnvironmentPath("musicconf"), "environments.ini") 'Dateipfad erstellen
        If Not File.Exists(f) Then Throw New FileNotFoundException("Die folgende Datei konnte nicht gefunden werden: " & Environment.NewLine & f) 'Datei existiert nicht
        environments = New List(Of MusicEnvironment) 'Liste initialisieren
        Dim text = String.Empty

        Using sr As StreamReader = File.OpenText(f) 'Datei öffnen

            'zeilenweise einlesen
            Do Until sr.EndOfStream
                text &= sr.ReadLine
            Loop

        End Using


        'Auswerten
        Dim parts() = text.Split({";"c}, StringSplitOptions.RemoveEmptyEntries) 'Inhalt unterteilen

        For i = 0 To parts.Length - 1 'Umgebungen iterieren

            Dim env As New MusicEnvironment(parts(i), i) 'erstellen
            environments.Add(env) 'zur Liste hinzufügen

        Next

    End Sub

    Private Sub New(name As String, ID As Integer)
        m_Name = name
        m_ID = ID
    End Sub

    Friend Shared Function getEnvironment(name As String) As MusicEnvironment
        For i = 0 To environments.Count - 1
            Dim env As MusicEnvironment = environments.Item(i)
            If env.name = name Then Return env
        Next
        Return Nothing
    End Function

    Friend Shared Function getEnvironment(ID As Integer) As MusicEnvironment
        For i = 0 To environments.Count - 1
            Dim env As MusicEnvironment = environments.Item(i)
            If env.ID = ID Then Return env
        Next
        Return Nothing
    End Function

    Public Shared Operator =(env1 As MusicEnvironment, env2 As MusicEnvironment) As Boolean
        If env1 Is Nothing OrElse env2 Is Nothing Then Return False
        Return env1.m_ID = env2.m_ID
    End Operator

    Public Shared Operator <>(env1 As MusicEnvironment, env2 As MusicEnvironment) As Boolean
        If env1 Is Nothing OrElse env2 Is Nothing Then Return False
        Return env1.m_ID <> env2.m_ID
    End Operator
End Class