﻿Option Strict On
Option Infer On

Imports System.Threading
Imports System.IO
Imports Helpers.Mathematics

Friend Class World
    Private entities As List(Of Entity)
    Private lastStep As DateTime
    Private exitGame As Boolean
    Private steps As Integer
    Private m_money As Integer
    Private m_life As Integer
    Private m_way As Way
    Private m_size As Size
    Private m_desc As String
    Private spawner As SpawnControl
    Private paused As Boolean
    Private custom As Boolean
    Private m_name As String
    Private m_speed As Double
    Private curControl As CursorControl
    Private startMoney As Integer
    Private m_gameRunsSlow As Boolean
    Private msgControl As MessageControl
    Private msgShown As Boolean
    Private msgText As String
    Private allowedTowers As UInteger

    Private Shared entLock As New Object

    Friend Event moneyChanged(sender As Object, e As EventArgs)
    Friend Event lifeChanged(sender As Object, e As EventArgs)
    Friend Event currentWaveChanged(sender As Object, e As EventArgs)
    Friend Event Won As EventHandler
    Friend Event Lost As EventHandler
    
    Friend Function getEntities() As Entity()
        SyncLock entLock
            Return entities.ToArray
        End SyncLock
    End Function

    Friend Property money As Integer
        Get
            Return m_money
        End Get
        Private Set(value As Integer)
            m_money = value
            RaiseEvent moneyChanged(Me, EventArgs.Empty)
        End Set
    End Property

    Friend Property life As Integer
        Get
            Return m_life
        End Get
        Private Set(value As Integer)
            m_life = value
            RaiseEvent lifeChanged(Me, EventArgs.Empty)
            If m_life <= 0 Then onLost()
        End Set
    End Property

    Friend ReadOnly Property way As Way
        Get
            Return m_way
        End Get
    End Property

    Friend ReadOnly Property size As Size
        Get
            Return m_size
        End Get
    End Property

    Friend ReadOnly Property isCustom As Boolean
        Get
            Return custom
        End Get
    End Property

    Friend ReadOnly Property name As String
        Get
            Return m_name
        End Get
    End Property

    Friend ReadOnly Property description As String
        Get
            Return m_desc
        End Get
    End Property

    Friend Property speed As Double
        Get
            Return m_speed
        End Get
        Set(value As Double)
            If value <= 0 Then Exit Property
            m_speed = value
        End Set
    End Property

    Friend ReadOnly Property waves As Integer
        Get
            Return spawner.waves
        End Get
    End Property

    Friend ReadOnly Property currentWave As Integer
        Get
            Return spawner.currentWave
        End Get
    End Property

    Friend ReadOnly Property GamePaused As Boolean
        Get
            Return paused
        End Get
    End Property

    Friend ReadOnly Property MessageVisible As Boolean
        Get
            Return msgShown
        End Get
    End Property

    Friend ReadOnly Property MessageText As String
        Get
            Return msgText
        End Get
    End Property

    Friend ReadOnly Property GameRunsSlow As Boolean
        Get
            Return m_gameRunsSlow
        End Get
    End Property

    Friend Sub HideMessage()
        msgShown = False
    End Sub

    Friend Function GetSteps() As Integer
        Return steps
    End Function

    Friend Function pay(val As Integer) As Boolean
        If money >= val Then
            money -= val
            Return True
        End If
        Return False
    End Function

    Friend Sub addMoney(val As Integer)
        money += val
    End Sub

    Friend Function IsTowerAllowed(index As Integer) As Boolean
        Dim towerInt = 1UI << index
        Return (allowedTowers And towerInt) = towerInt
    End Function

    Friend Function canBuild(rect As RectangleF) As Boolean
        Dim ents() = getEntities()
        For Each ent In ents
            If TypeOf ent Is Tower Then
                Do While ent.InUse
                    Thread.Sleep(1)
                Loop
                ent.InUse = True
                If rect.IntersectsWith(ent.getRect) Then
                    ent.InUse = False
                    Return False
                End If
                ent.InUse = False
            End If
        Next
        Dim checkPoints(7) As Vector2
        checkPoints(0) = New Vector2(rect.Location)
        checkPoints(1) = New Vector2(rect.X + rect.Width / 2.0F, rect.Y)
        checkPoints(2) = New Vector2(rect.X + rect.Width, rect.Y)
        checkPoints(3) = New Vector2(rect.X + rect.Width, rect.Y + rect.Height / 2.0F)
        checkPoints(4) = New Vector2(rect.X + rect.Width, rect.Y + rect.Height)
        checkPoints(5) = New Vector2(rect.X + rect.Width / 2.0F, rect.Y + rect.Height)
        checkPoints(6) = New Vector2(rect.X, rect.Y + rect.Height)
        checkPoints(7) = New Vector2(rect.X, rect.Y + rect.Height / 2.0F)
        For i = 0 To checkPoints.Length - 1
            If checkPoints(i).X < 0 OrElse checkPoints(i).Y < 0 OrElse checkPoints(i).X > size.Width OrElse checkPoints(i).Y > size.Height Then Return False
        Next
        For i = 0 To way.segments.Length - 1
            Dim points() = way.segments(i).GetPoints(5)
            For j = 0 To points.Length - 1
                Dim m As New Vector2(points(j))
                For k = 0 To checkPoints.Length - 1
                    If MathHelper.pointInCircle(m, 14, checkPoints(k)) Then Return False
                Next
            Next
        Next
        Dim lastP As New Vector2(way.segments.Last.p2)
        For i = 0 To checkPoints.Length - 1
            If MathHelper.pointInCircle(lastP, 20, checkPoints(i)) Then Return False
        Next
        Return True
    End Function

    Friend Sub reduceHealth()
        life -= 1
    End Sub

    Private Sub onLost()
        pause()
        RaiseEvent Lost(Me, EventArgs.Empty)
    End Sub

    Sub New(name As String, type As MapType)
        m_name = name
        m_speed = 60.0
        entities = New List(Of Entity)
        spawner = New SpawnControl(Me)
        AddHandler spawner.beginSpawn, Sub() RaiseEvent currentWaveChanged(Me, EventArgs.Empty)
        Dim mapPath = String.Empty
        Select Case type
            Case MapType.Tutorial
                mapPath = Path.Combine(Game.getEnvironmentPath("tutorialmaps"), name)
            Case MapType.Campaign
                mapPath = Path.Combine(Game.getEnvironmentPath("campaignmaps"), name)
            Case Else
                mapPath = Path.Combine(Game.getEnvironmentPath("maps"), name)
        End Select
        allowedTowers = &HFFFFFFFFUI
        loadWorld(mapPath)
        msgControl = New MessageControl(mapPath)
    End Sub

    Friend Sub start(life As Integer, steps As Integer, control As WorldControl)
        Me.life = life
        Me.money = startMoney
        Me.steps = steps
        curControl = New CursorControl(Me, control)
        Dim ents() = getEntities()
        For i = 0 To ents.Length - 1
            Do While ents(i).InUse
                Thread.Sleep(1)
            Loop
            ents(i).InUse = True
            despawnEntity(ents(i))
            ents(i).InUse = False
        Next
        spawner.reset()
        exitGame = False
        paused = False
        msgShown = False
        Dim t As New Thread(New ThreadStart(AddressOf ExtraLoop))
        t.Name = "second game thread"
        t.IsBackground = True
        t.Start()
        UpdateLoop()
    End Sub

    Private Sub loadWorld(path As String)
        Dim f = IO.Path.Combine(path, "map.td1")
        Dim fs = File.Open(f, FileMode.Open, FileAccess.Read)
        Dim br As New BinaryReader(fs)
        Dim bytes(Convert.ToInt32(fs.Length) - 1) As Byte
        For i = 0 To bytes.Length - 1
            bytes(i) = br.ReadByte
        Next
        br.Close()
        br.Dispose()
        fs.Close()
        fs.Dispose()
        Dim seperator As Byte = &H7F
        Dim valueSets As New List(Of Byte())
        Dim index = 0
        Do While index < bytes.Length
            Dim bList As New List(Of Byte)
            Do While index < bytes.Length AndAlso Not bytes(index) = seperator
                bList.Add(bytes(index))
                index += 1
            Loop
            valueSets.Add(bList.ToArray)
            index += 1
        Loop
        For i = 0 To valueSets.Count - 1
            parseBytes(valueSets.Item(i))
        Next
        m_way = New Way(path, Me)
    End Sub

    Private Sub parseBytes(ByVal vals() As Byte)
        Dim index = vals(0)
        Select Case index
            Case 0
                Exit Sub
            Case 1
                m_size.Width = BitConverter.ToInt32(vals, 1)
                m_size.Height = BitConverter.ToInt32(vals, 5)
            Case 2
                Dim chars() = System.Text.Encoding.UTF8.GetChars(vals, 1, vals.Length - 1)
                m_desc = New String(chars)
            Case 3
                Dim sI As New SpawnInfo
                sI.time = BitConverter.ToInt32(vals, 1)
                Dim spawns(vals.Length - 6) As Byte
                Array.Copy(vals, 5, spawns, 0, spawns.Length)
                sI.spawns = spawns
                spawner.addSpawn(sI)
            Case 4
                custom = BitConverter.ToBoolean(vals, 1)
            Case 5
                startMoney = BitConverter.ToInt32(vals, 1)
            Case 6
                allowedTowers = BitConverter.ToUInt32(vals, 1)
        End Select
    End Sub

    Friend Sub spawnEntity(e As Entity)
        SyncLock entLock
            entities.Add(e)
            e.world = Me
            e.onSpawned()
        End SyncLock
    End Sub

    Friend Sub despawnEntity(e As Entity)
        SyncLock entLock
            entities.Remove(e)
            e.world = Nothing
            e.onDespawned()
        End SyncLock
    End Sub

    Private Sub onStep()
        Dim ents() = getEntities()
        For i = 0 To ents.Length - 1
            Do While ents(i).InUse
                Thread.Sleep(1)
            Loop
            ents(i).InUse = True
            ents(i).doStep()
            ents(i).InUse = False
        Next
    End Sub

    Private Sub UpdateLoop()
        Dim sw As New Stopwatch
        Dim runsSlow = False
        Dim elapsed = 0L
        Do Until exitGame
            If paused Then
                Thread.Sleep(1)
                Continue Do
            End If
            If msgShown Then
                Thread.Sleep(1)
                Continue Do
            End If
            sw.Start()
            steps += 1
            spawner.doStep()
            onStep()
            Dim msg = String.Empty
            If msgControl.CheckForMessage(steps, msg) Then
                msgText = msg
                msgShown = True
            End If
            sw.Stop()
            elapsed = sw.ElapsedMilliseconds
            sw.Reset()
            Dim targetTime = 1000 / speed
            If elapsed < targetTime Then
                runsSlow = False
                Thread.Sleep(Convert.ToInt32(targetTime - elapsed))
            Else
                runsSlow = True
            End If
            m_gameRunsSlow = runsSlow
        Loop
        curControl = Nothing
    End Sub

    Private Sub ExtraLoop()
        Do Until exitGame
            If paused Then
                Thread.Sleep(1)
                Continue Do
            End If
            curControl.checkForCursorChanges()
            If HasWon() Then
                RaiseEvent Won(Me, EventArgs.Empty)
                pause()
                Continue Do
            End If
            Thread.Sleep(1)
        Loop
    End Sub

    Private Function HasWon() As Boolean
        If currentWave = waves AndAlso Not spawner.IsSpawning Then
            Dim ents = getEntities()
            For Each ent In ents
                If TypeOf ent Is Creep Then Return False
            Next
            Return True
        Else
            Return False
        End If
    End Function

    Friend Sub pause()
        paused = True
    End Sub

    Friend Sub [resume]()
        paused = False
    End Sub

    Friend Sub [stop]()
        exitGame = True
    End Sub

    Private Class CursorControl
        Private world As World
        Private control As WorldControl

        Private Delegate Function pointToClient(p As Point) As Point
        Private Delegate Sub changeCursor(i As Integer)

        Sub New(w As World, c As WorldControl)
            world = w
            control = c
        End Sub

        Friend Sub checkForCursorChanges()
            Dim changeCurDel As New changeCursor(AddressOf Game.window.ChangeCursor)
            If world.exitGame Then
                Game.window.BeginInvoke(changeCurDel, 0)
                Return
            End If
            Dim offset = control.getOffset
            Dim curPos = DirectCast(Game.window.Invoke(New pointToClient(AddressOf control.PointToClient), Cursor.Position), Point)
            curPos.X -= offset.X
            curPos.Y -= offset.Y
            Dim ents() = world.getEntities
            Dim cursorChanged = False

            If control.isBuilding Then
                Game.window.BeginInvoke(changeCurDel, 3)
                cursorChanged = True
            Else
                For Each ent In ents
                    Do While ent.InUse
                        Thread.Sleep(1)
                    Loop
                    ent.InUse = True
                    If ent.getRect.Contains(curPos) Then
                        If TypeOf ent Is Tower Then
                            Game.window.BeginInvoke(changeCurDel, 1)
                            cursorChanged = True
                            ent.InUse = False
                            Exit For
                        ElseIf TypeOf ent Is Creep AndAlso control.selectedTower IsNot Nothing Then
                            Game.window.BeginInvoke(changeCurDel, 2)
                            cursorChanged = True
                            ent.InUse = False
                            Exit For
                        End If
                    End If
                    ent.InUse = False
                Next
            End If
            If Not cursorChanged Then Game.window.BeginInvoke(changeCurDel, 0)
        End Sub
    End Class
End Class

Friend Structure SpawnInfo
    Friend time As Integer
    Friend spawns() As Byte
End Structure

Friend Class SpawnControl
    Private world As World
    Private spawns As List(Of SpawnInfo)
    Private steps As Integer
    Private spawnIndex As Integer
    Private stepsToWait As Integer
    Private creepsSpawned As Integer
    Private spawning As Boolean

    Friend Event beginSpawn(sender As Object, e As EventArgs)
    Friend Event endSpawn(sender As Object, e As EventArgs)

    Sub New(w As World)
        world = w
        spawns = New List(Of SpawnInfo)
        reset()
    End Sub

    Friend ReadOnly Property waves As Integer
        Get
            Return spawns.Count
        End Get
    End Property

    Friend ReadOnly Property currentWave As Integer
        Get
            Return spawnIndex + 1
        End Get
    End Property

    Friend ReadOnly Property IsSpawning As Boolean
        Get
            Return spawning
        End Get
    End Property

    Friend Sub reset()
        steps = 0
        spawnIndex = -1
        stepsToWait = 0
        creepsSpawned = 0
        spawning = False
    End Sub

    Friend Sub addSpawn(info As SpawnInfo)
        spawns.Add(info)
    End Sub

    Friend Sub doStep()
        If spawning Then stepsToWait -= 1
        checkSpawns()
        steps += 1
    End Sub

    Private Sub checkSpawns()
        For i = 0 To spawns.Count - 1
            If spawns.Item(i).time = steps Then
                spawnIndex = i
                stepsToWait = 0
                creepsSpawned = 0
                spawning = True
                RaiseEvent beginSpawn(Me, EventArgs.Empty)
                Exit For
            End If
        Next
        If spawning AndAlso stepsToWait <= 0 Then
            spawnCreep(spawnIndex, creepsSpawned)
            creepsSpawned += 1
            If creepsSpawned >= getCreepCount(spawnIndex) Then
                spawning = False
                RaiseEvent endSpawn(Me, EventArgs.Empty)
            End If
        End If
    End Sub

    Private Sub spawnCreep(sIndex As Integer, cIndex As Integer)
        Dim creeps(getCreepCount(sIndex) - 1) As Integer
        Dim counter = 0
        For i = 0 To spawns.Item(sIndex).spawns.Length - 1 Step 5
            For j = 1 To spawns.Item(sIndex).spawns(i + 4)
                creeps(counter) = BitConverter.ToInt32(spawns.Item(sIndex).spawns, i)
                counter += 1
            Next
        Next
        Dim creepType = creeps(cIndex)
        Dim c As Creep = Nothing
        For Each desc In Creep.GetAllCreeps
            If desc.ID = creepType Then
                c = DirectCast(Activator.CreateInstance(desc.CreepType), Creep)
                Exit For
            End If
        Next
        stepsToWait = Convert.ToInt32(10 / c.speed)
        world.spawnEntity(c)
    End Sub

    Private Function getCreepCount(sIndex As Integer) As Integer
        Dim count = 0
        For i = 0 To spawns.Item(sIndex).spawns.Length - 1 Step 5
            count += spawns.Item(sIndex).spawns(i + 4)
        Next
        Return count
    End Function
End Class

Friend Class MessageControl
    Private messages As List(Of Message)

    Friend Function CheckForMessage(time As Integer, ByRef messageText As String) As Boolean
        Dim found = From m In messages Where m.Time = time Select m.Text
        If found.Count > 0 Then
            messageText = found.First
            Return True
        End If
        Return False
    End Function

    Sub New(path As String)
        messages = New List(Of Message)
        Dim fi As New FileInfo(IO.Path.Combine(path, "map.td3"))
        If fi.Exists Then
            Using fs = fi.OpenRead
                Using sr As New StreamReader(fs)
                    'Debug.Print("MessageControl: Filestream open, begin reading message file.")
                    Do Until sr.EndOfStream
                        Dim messageOpen = False
                        Dim messageTime = 0
                        Dim messageText = String.Empty
                        Dim messageValid = False
                        Do Until messageOpen OrElse sr.EndOfStream
                            Dim line = sr.ReadLine.ToLower.Trim
                            If line.StartsWith("[message:") AndAlso line.EndsWith("]") AndAlso Integer.TryParse(line.Substring(9, line.Length - 10), messageTime) Then
                                'Debug.Print("MessageControl: Begin of Message detected, opening message.")
                                messageOpen = True
                            Else
                                'Debug.Print("MessageControl: Ignoring line, no message open.")
                            End If
                        Loop
                        Do While messageOpen AndAlso Not sr.EndOfStream
                            Dim line = sr.ReadLine.Trim
                            If String.IsNullOrWhiteSpace(line) Then
                                Debug.Print("MessageControl: Line contains only whitespace, skipping.")
                                Continue Do
                            End If
                            Dim blockOpen = False
                            If line.StartsWith("{") Then
                                'Debug.Print("MessageControl: Begin of message block found, start read.")
                                If line.Length > 1 Then messageText = line.Substring(1)
                                blockOpen = True
                            Else
                                messageOpen = False
                                'Debug.Print("MessageControl: Invalid syntax found while reading message, aborting.")
                            End If
                            Do While blockOpen AndAlso Not sr.EndOfStream
                                Dim l = sr.ReadLine.Trim
                                Dim i = l.IndexOf("}")
                                If i > -1 Then
                                    'Debug.Print("MessageControl: End of message block found, closing message.")
                                    If i > 0 Then messageText &= Environment.NewLine & l.Substring(0, i)
                                    blockOpen = False
                                    messageOpen = False
                                Else
                                    If Not String.IsNullOrWhiteSpace(messageText) Then messageText &= Environment.NewLine
                                    messageText &= l
                                End If
                            Loop
                            If Not blockOpen Then messageValid = True
                        Loop
                        If messageValid Then
                            'Debug.Print("MessageControl: Message reading complete, adding message.")
                            messages.Add(New Message(messageTime, messageText))
                        End If
                    Loop
                End Using
            End Using
        End If
        'messages.ForEach(Sub(m) Debug.Print(String.Join(" ", m.Time, ":", m.Text)))
    End Sub

    Private Class Message
        Private m_Time As Integer
        Private m_Text As String

        Friend ReadOnly Property Time As Integer
            Get
                Return m_Time
            End Get
        End Property

        Friend ReadOnly Property Text As String
            Get
                Return m_Text
            End Get
        End Property

        Friend Sub New(time As Integer, text As String)
            m_Time = time
            m_Text = text
        End Sub
    End Class
End Class

Friend Enum MapType
    SinglePlayer
    Tutorial
    Campaign
End Enum